/* HtmlFilter.m -- Html filtering and description services for DL indexing
	of  Html docs.
	The following two services are implemented.

[1] HtmlDescribe Service:
	Describes Html articles based currently on title, h1, h2 and h3 tags.
	This can be modified quite easily.  Some effort has been made to extract
	full name of the poster from the From: line.

	This description service is advertised via the following specification in
	~/Library/Services/HtmlIndexing.service/services (or /LocalLibrary...) file.
		Filter: HtmlDescribe
		Port: IXHTMLFILTER
		Send Type: NXTypedFileContentsPboardType:html
		Return Type: IXFileDescriptionPboardType
		Executable: HtmlFilter.daemon

[2] HtmlFilter Service:
	Filters junk out of Html articles, so junk will not be indexed.
	"Junk" currently is defined as:
	    tag keywords
	    href tags

	This filtering service is advertised via the following specification in
	~/Library/Services/HtmlIndexing.service/services (or /LocalLibrary...) file.
		Filter: HtmlFilter
		Port: IXHTMLFILTER
		Send Type: NXTypedFileContentsPboardType:html
		Return Type: NXAsciiPboardType
		Executable: HtmlFilter.daemon

	Advantage of this service daemon scheme over the Unix stdio filter
	(invoked via NXUNIXSTDIO port) is that the daemon based filter is
	invoked only once per DL indexing session, unlike stdio filter which
	is invoked for every article indexed.

	Daemons can keep running indefinitely, but this one quits after some
	duration of inactivity.

	No Copyright is claimed.
	This program is hereby released into the public domain.

	Juergen Sell js@euler.han.de.

------- TO DO ----------------------------------------------------------------
[1] Making things configurable, e.g., what goes into the description line in
	what order, and what kind of lines to filter out in -HtmlFilter:...
	method.  These specs can be stored in a file in
	~/Library/Services/HtmlIndexing.service directory (or /LocalLibrary..)
	This path should be in argv[0] in main() because that is where the
	executable lives too.
	I probably won't get around to doing this myself, or doing much
	enhancements.

------- Revision History -----------------------------------------------------
98//11/26
	Version 0.13 of HtmlDescribe. Juergen Sell js@euler.han.de
	Stolen from HewsIndexing,Izumi Ohzawa, izumi@pinoko.berkeley.edu 
	Errors are mine.
98//11/29
	Version 0.2 of HtmlDescribe. Juergen Sell js@euler.han.de
	treating entire data stream at once per regexps as apposed to 
	on a per line basis. That was inappropriate for html.
98//12/03
	Version 0.21 of HtmlDescribe. Juergen Sell js@euler.han.de
	eliminate some memory leaks.
-------------------------------------------------------------------------------
*/


#import <appkit/appkit.h>
#import <indexing/indexing.h>

#import <stdio.h>
#import <stdlib.h>
#import <strings.h>
#import <time.h>

#include <misckit.h>

// Uncomment the following during debugging.  See Console output.
#define MYDEBUG 1

static char *version = "[V0.21, Juergen Sell js@euler.han.de]";

// Timeout stuff
static  float timeoutvalue = 120.0;	// wait this long before quitting.
static  float interval = 30.0;		// timeout check interval
static  time_t timelast;		// last time (in seconds) that this service is called.

// The following must match the spec in "services" file.
char *InPBType = "NXTypedFileContentsPboardType:html";

#define DESCRIPTION_LENGTH 1000
@interface Provider:Object
{
    DPSTimedEntry	timer;
}
- init;
- step;
- HtmlDescribe:(id)pbid userData:(const char *)udata error:(char **)errmsg;
- HtmlFilter:(id)pbid userData:(const char *)udata error:(char **)errmsg;
@end



@implementation Provider

void runOneStep(DPSTimedEntry timedEntry, double timeNow, void *TEobject)
{
    [(id)TEobject step];
}


- init
{
    time(&timelast);		// initialize no-work timer.
    timer = DPSAddTimedEntry(interval, &runOneStep, self, NX_RUNMODALTHRESHOLD);
    [super init];
    return self;
}

- step
{
time_t timenow;
    time(&timenow);
    if( (timenow - timelast) >= timeoutvalue) {
	DPSRemoveTimedEntry (timer);
	fprintf(stderr,
		"HtmlFilter: No requests received for %.1f seconds. Quitting: %s",
		timeoutvalue, ctime(&timenow));
	_exit(2);
    }
    return self;
}


#define NDESCKEY	  4		/* # of names in *desckwords[] */

// Key words for news description extraction.
static const char *desckwords[] = {
	"<TITLE>.*</TITLE>",    "<H1>.*</H1>",  
	"<H2>.*</H2>",          "<H3>.*</H3>",
	""
    };

// The main work-horse method to extract description line based on
// Subject:, From:, and Date: lines from data passed via pasteboard.
//
- HtmlDescribe:(id)pbid userData:(const char *)udata error:(char **)errmsg
{
char *htmldata; 
int  length;
int  i,r;
id pmString= [[MiscString alloc] init],
  lineString,flineString=[MiscString alloc] ,
  resString= [[MiscString alloc] init];

    	time(&timelast);		// refresh the no-work timer.

 	if([pbid findAvailableTypeFrom:&InPBType num:1]) {
	  // We have correct PB type.
	  [pbid readType:InPBType data:&htmldata length:&length];
	  if(htmldata && length) {
	    // PB has some actual data in it.

	    [flineString initString:htmldata];
	    lineString= [flineString left:DESCRIPTION_LENGTH];
	    [flineString free];
	    [lineString replaceEveryOccurrenceOfChar:'\n' with:""];
	    [lineString replaceEveryOccurrenceOfChar:'\r' with:""];
#if MYDEBUG>1
	    fprintf(stderr, "line=  =%s=\n", [lineString stringValue]);
#endif
	    for(i=0; i<NDESCKEY; i++) {
	      r= [lineString grep: desckwords[i]
			     caseSensitive:NO
			     before:nil 
			     middle:pmString 
			     after:nil];
	      if (r == 1) {
		if (i > 0) [resString cat:" -- "];
		[resString concatenate:pmString];
	      }
#if MYDEBUG>1
	      fprintf(stderr, "match=%d  =%s=\n", r, [pmString stringValue]);
#endif
	    }   

	    [resString replaceEveryOccurrenceOfRegex:"<[^>]*>" with:""];
	    [resString squashSpaces];
	    // Send the description back on pasteboard.
	    [pbid declareTypes:&IXFileDescriptionPboardType num:1 owner:self];
	    [pbid writeType:IXFileDescriptionPboardType data:[resString stringValue]
		  length:[resString length]];
#ifdef MYDEBUG
		fprintf(stderr, ">%s\n", [resString stringValue]);
#endif
	    } /* end if(htmldata && length) */
	    else
		*errmsg = "No actual data found on pasteboard.";

 	    [pbid deallocatePasteboardData:htmldata length:length];
	}  /* end if(types[i]) */
	else
	    *errmsg = "No good PBoard type.";

	[lineString free];
	[pmString free];
	[resString free];

	return self;
}


// These skip and done regexp are hard coded as below currently.
// Ideally, these should be modifiable via a word list file included in
// ~/Library/Services/NewsIndexing.service directory (or /LocalLibrary...)
// Path to this directory should be obtainable via argv[0] in main().
// I am not going to do this, so if there are any takers, please!
//
static const char *lskipwords[] = {
	"<A>.*</A>",   "<[^>]*>",  
	""
    };
#define NLSKIP	  2		/* # of names in *lskipwords[] */

// Html article filter service method that removes junk (see top of this file)
// from article text before it is passed to indexing scanner.
// DigitalLibrarian calles this method with full article content on the
// pasteboard of type "NXTypedFileContentsPboardType:html".
// This method is called before the description filter service 
// -HtmlDescribe:..... method is called.
//
- HtmlFilter:(id)pbid userData:(const char *)udata error:(char **)errmsg
{

char *htmldata; 
int  length;
int  i;
id lString= [MiscString alloc];

    	time(&timelast);		// refresh the no-work timer.

 	if([pbid findAvailableTypeFrom:&InPBType num:1]) {
	    // We have correct PB type.
	    [pbid readType:InPBType data:&htmldata length:&length];
	    if(htmldata && length) {
		// PB has some actual data in it.
		// Some initializations.
		[lString initCapacity:length];
		[lString setStringValue:htmldata];
		for(i=0; i<NLSKIP; i++) {
		  [lString replaceEveryOccurrenceOfRegex:lskipwords[i]
			      with:""];
		}
#if MYDEBUG==3
		fprintf(stderr, "+>> %s\n", [lString stringValue]);
#endif

		// Send the description back on pasteboard.
 		[pbid declareTypes:&NXAsciiPboardType num:1 owner:self];
 		[pbid writeType:NXAsciiPboardType data:[lString stringValue] 
		      length:[lString length]];
	    } /* end if(htmldata && length) */
	    else
		*errmsg = "No actual data found on pasteboard.";

 	    [pbid deallocatePasteboardData:htmldata length:length];
	}  /* end if(types[i]) */
	else
	    *errmsg = "No good PBoard type.";
	[lString free];
	return self;
}

@end


int main(int argc, char *argv[])
{  
id lid;
#ifdef MYDEBUG
int i;
#endif

    time(&timelast);
    fprintf(stderr,
	"HtmlFilter.deamon for DLibrarian indexing started: %s %s\n",
			ctime(&timelast), version);

#ifdef MYDEBUG
    for(i=0; i<argc; i++)
	fprintf(stderr, "argv[%d]: %s\n", i, argv[i]);
#endif

    lid = [[Listener alloc] init];
    [lid setServicesDelegate:[[Provider alloc] init]];
    if([lid checkInAs:"IXHTMLFILTER"]) _exit(1);	// uh-oh!
    [lid addPort];
    [Listener run];
    return 0;
}
